"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.postpromorphism = exports.prepromorphism = exports.zygomorphism = exports.hylomorphism = exports.apomorphism = exports.anamorphism = exports.catamorphism = undefined;
exports.cata = cata;
exports.ana = ana;
exports.para = para;
exports.apo = apo;
exports.hylo = hylo;
exports.zygo = zygo;
exports.gApo = gApo;
exports.prepro = prepro;
exports.postPro = postPro;

var _Either = require("flow-static-land/lib/Either");

var _Fix = require("./Fix");

function cata(transformer, term, functor) {
  var extracted = (0, _Fix.out)(term),
      childrenMapped = functor.map(function (x) {
    return cata(transformer, x, functor);
  }, extracted),
      transformed = transformer(childrenMapped);

  return transformed;
}

var catamorphism = exports.catamorphism = cata;

function ana(transformer, seed, functor) {
  var transformed = transformer(seed),
      childrenMapped = functor.map(function (x) {
    return ana(transformer, x, functor);
  }, transformed),
      rewrapped = new _Fix.In(childrenMapped);

  return rewrapped;
}
var anamorphism = exports.anamorphism = ana;

function para(transformer, term, functor) {
  var extracted = (0, _Fix.out)(term),
      childrenMapped = functor.map(function (x) {
    return para(transformer, x, functor);
  }, extracted),
      transformed = transformer(term, childrenMapped);

  return transformed;
}

function id(a) {
  return a;
}

function apo(transformer, seed, functor) {

  var transformed = transformer(seed);

  function fanIn(e) {
    return (0, _Either.either)(id, function (x) {
      return apo(transformer, x, functor);
    }, e);
  }

  var childrenMapped = functor.map(fanIn, transformed);
  return new _Fix.In(childrenMapped);
}
var apomorphism = exports.apomorphism = apo;

function hylo(tearDown, buildUp, seed, functor) {
  var built = buildUp(seed),
      recursed = functor.map(function (x) {
    return hylo(tearDown, buildUp, x, functor);
  }, built),
      flattened = tearDown(recursed);

  return flattened;
}
var hylomorphism = exports.hylomorphism = hylo;

function zygoAlgebra(alg, almostAlgebra, functor) {
  return function (wrappedPair) {
    var secondWrapped = functor.map(function (pair) {
      return pair[1];
    }, wrappedPair),
        secondReduced = alg(secondWrapped),
        firstReduced = almostAlgebra(wrappedPair);

    return [firstReduced, secondReduced];
  };
}

function zygo(tearDown, coalesce, term, functor) {
  return cata(zygoAlgebra(tearDown, coalesce, functor), term, functor)[0];
};
var zygomorphism = exports.zygomorphism = zygo;

function gApo(expand, expandA, seed, functor) {

  var expanded = expandA(seed);

  function fanIn(e) {
    return (0, _Either.either)(function (a) {
      return gApo(expand, expandA, a, functor);
    }, function (b) {
      return ana(expand, b, functor);
    }, e);
  }

  var transformed = functor.map(fanIn, expanded);
  return new _Fix.In(transformed);
};

function prepro(naturalTransformer, transformer, term, functor) {
  var extracted = (0, _Fix.out)(term),
      childrenMapped = functor.map(function (x) {
    return prepro(naturalTransformer, transformer, x, functor);
  }, extracted),
      childrenMappedTransformed = naturalTransformer(childrenMapped),
      transformed = transformer(childrenMappedTransformed);

  return transformed;
}
var prepromorphism = exports.prepromorphism = prepro;

function postPro(naturalTransformer, expand, seed, functor) {
  var expanded = expand(seed),
      naturalTransformed = naturalTransformer(expanded),
      childrenMapped = functor.map(function (x) {
    return postPro(naturalTransformer, expand, x, functor);
  }, naturalTransformed),
      rewrapped = new _Fix.In(childrenMapped);

  return rewrapped;
}
var postpromorphism = exports.postpromorphism = postPro;